/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//(C) PicoPeta Simputers Pvt. Ltd. (2003)
// Image Library

// Programmer: Santhosh N

#include "pimg_gif.h"

int load_gif_image(FILE *fp, Display *disp, struct Image **simg, unsigned short iattr)
{
	GifFileType *gif;
	GifRecordType rec;
	GifPixelType *row=NULL;
	ColorMapObject *cmap;
	int transpval=-1, x, y, ioff[]={0,4,2,1}, ijmp[]={8,8,4,2}, tdisp;
	unsigned int t, il, it, iw, ih, i, niter, cn, delay=0, depth, mbytes, mbpl=0;
	unsigned char r, g, b, *ibuff=NULL, *imask=NULL;
	unsigned long p1, p2;
	struct Image *img=NULL;
	unsigned char bm[8]={0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80};

	depth = DefaultDepth(disp, DefaultScreen(disp));

	gif = DGifOpenFileHandle(fileno(fp));

	if(gif==NULL)
		return EBADIMG;

	if(gif->SWidth<1 || gif->SHeight<1) {
		DGifCloseFile(gif);
		return EBADIMG;
	}

	// Process GIF records one by one
	do {
		if(DGifGetRecordType(gif, &rec)==GIF_ERROR) {
			DGifCloseFile(gif);
			return EBADIMG;
		}

		if(rec==IMAGE_DESC_RECORD_TYPE) {
			if(DGifGetImageDesc(gif)==GIF_ERROR) {
				DGifCloseFile(gif);
				return EBADIMG;
			}

			iw = gif->Image.Width;
			ih = gif->Image.Height;
			il = gif->Image.Left;
			it = gif->Image.Top;

			// Compare the image dimension for maximum limit
			if(iw+il > gif->SWidth || ih+it > gif->SHeight || iw*ih > MAX_WIDTH*MAX_HEIGHT) {
				DGifCloseFile(gif);
				return EBADIMG;
			}

			// Get the colormap
			cmap = (gif->Image.ColorMap)?gif->Image.ColorMap:gif->SColorMap;

			if(row)
				free(row);
			row = malloc(iw*sizeof(GifColorType));
			if(row==NULL) {
				DGifCloseFile(gif);
				return ENOMEMR;
			}

			// Allocate memory for the image
			if(depth==4) {
				ibuff = (unsigned char*)malloc(iw*(ih+1));
			} else {
				ibuff = (unsigned char*)malloc(iw*(ih+1)*2);
			}
			if(ibuff==NULL) {
				free(row);
				DGifCloseFile(gif);
				return ENOMEMR;
			}
	
			// Allocate memory for the image mask
			if(iattr & SHAPE_MASK) {
				mbpl = iw/8+((iw%8)?1:0);
				mbytes = mbpl*ih;
				imask = (unsigned char*)malloc(mbytes);
				memset(imask, 0x00, mbytes);
			}

			// Create a node
			img = (struct Image*)malloc(sizeof(struct Image));
			img->next = NULL;
			img->delay = delay;
			img->x = il;
			img->y = it;
			img->width = iw;
			img->height = ih;
			img->img = img->mimg = NULL;

			if(gif->Image.Interlace) {
				niter = 4;
			} else {
				niter = 1;
				ioff[0] = 0;
				ijmp[0] = 1;
			}

			// Start reading the image data
			switch(depth) {
				case 4:
					for(i=0; i<niter; i++) {
						for(y=ioff[i]; y<ih; y+=ijmp[i]) {
							if(DGifGetLine(gif, row, iw)==GIF_ERROR) {
								free(row);
								free(img);
								free(ibuff);
								if(iattr & SHAPE_MASK)
									free(imask);
								DGifCloseFile(gif);
								return EBADIMG;
							}
							for(x=0; x<iw; x+=2) {
								cn = row[x];
								r=cmap->Colors[cn].Red; g=cmap->Colors[cn].Green;	b=cmap->Colors[cn].Blue;
								p1 = r*11 + g*16 + b*5;
								// Write to the shape mask
								if((iattr&SHAPE_MASK) && cn!=transpval)
							 		*(imask + y*mbpl + x/8) |= bm[x%8];

								cn = row[x+1];
								r=cmap->Colors[cn].Red; g=cmap->Colors[cn].Green;	b=cmap->Colors[cn].Blue;
								p2 = r*11 + g*16 + b*5;
								// Write to the shape mask
								if((iattr&SHAPE_MASK) && cn!=transpval)
							 		*(imask + y*mbpl + (x+1)/8) |= bm[(x+1)%8];

								p1 = (((p2>>9)<<4)|(p1>>9))&0xff;
								memcpy(ibuff + y*iw + x/2, &p1, 1);
							}
						}
					}
					break;
				case 12:
					 for(i=0; i<niter; i++) {
						 for(y=ioff[i]; y<ih; y+=ijmp[i]) {
							 if(DGifGetLine(gif, row, iw)==GIF_ERROR) {
								 free(row);
								 free(img);
								 free(ibuff);
								 if(iattr & SHAPE_MASK)
									free(imask);
								 DGifCloseFile(gif);
								 return EBADIMG;
							 }
							 for(x=0; x<iw; x++) {
								 cn = row[x];
								 r=cmap->Colors[cn].Red; g=cmap->Colors[cn].Green;	b=cmap->Colors[cn].Blue;
								 t = ((r>>4)<<8) + ((g>>4)<<4) + (b>>4);
								 memcpy(ibuff + y*iw*2 + x*2, &t, 2);
								
								 // Write to the shape mask
								 if((iattr&SHAPE_MASK) && cn!=transpval)
										 *(imask + y*mbpl + x/8) |= bm[x%8];
							 }
						 }
					 }
					 break;
				case 16:
					for(i=0; i<niter; i++) {
						for(y=ioff[i]; y<ih; y+=ijmp[i]) {
							if(DGifGetLine(gif, row, iw)==GIF_ERROR) {
								 free(row);
								 free(img);
								 free(ibuff);
								 if(iattr & SHAPE_MASK)
									free(imask);
								 DGifCloseFile(gif);
								 return EBADIMG;
							}
					
							if(_rmask==0x00f800 && _gmask==0x0007e0 && _bmask==0x00001f) {
								for(x=0; x<iw; x++) {
									cn = row[x];
									r=cmap->Colors[cn].Red; g=cmap->Colors[cn].Green; b=cmap->Colors[cn].Blue;
									
									t = ((r>>3)<<11) + ((g>>2)<<5) + (b>>3);
									
									memcpy(ibuff + y*iw*2 + x*2, &t, 2);
								    
									// Write to the shape mask
									if((iattr&SHAPE_MASK) && cn!=transpval)
										*(imask + y*mbpl + x/8) |= bm[x%8];
								}
							} else {
								for(x=0; x<iw; x++) {
									cn = row[x];
									r=cmap->Colors[cn].Red; g=cmap->Colors[cn].Green; b=cmap->Colors[cn].Blue;
									
									t = ((r>>2)<<10) + ((g>>3)<<5) + (b>>3);
									
									memcpy(ibuff + y*iw*2 + x*2, &t, 2);
								    
									// Write to the shape mask
									if((iattr&SHAPE_MASK) && cn!=transpval)
										*(imask + y*mbpl + x/8) |= bm[x%8];
								}
							}
						}
					}
					break;
			}

			// Add the Image node to the list
			if(depth==4) {
				img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								ibuff, img->width, img->height, 8, img->width);
			} else {
				img->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
								ibuff, img->width, img->height, 16, img->width*2);
			}
			
			// If image shape mask required, put it
			if(iattr & SHAPE_MASK) {
				img->mimg = XCreateImage(disp, CopyFromParent, 1, ZPixmap, 0,\
                                        imask, img->width, img->height, 8, mbpl);
				switch(depth) {
					case 4:
						img->mimg->red_mask = img->mimg->green_mask = \
											img->mimg->blue_mask = 0;
						break;
					case 12:
						img->mimg->red_mask = 3840;
						img->mimg->green_mask = 240;
						img->mimg->blue_mask = 15;
						break;
					case 16:
						if(_rmask==0x00f800 && _gmask==0x0007e0 && _bmask==0x00001f) {
							img->mimg->red_mask = 63488;
							img->mimg->green_mask = 2016;
							img->mimg->blue_mask = 31;
						} else {
							img->mimg->red_mask = 64512;
							img->mimg->green_mask = 992;
							img->mimg->blue_mask = 31;
						}
						break;
				}
			}
			
			_pimg_imglst_add(simg, img);
		} else if(rec==EXTENSION_RECORD_TYPE) {
			int ecode;
			GifByteType *ext=NULL;

			DGifGetExtension(gif, &ecode, &ext);
			if(ecode==GRAPHICS_EXT_FUNC_CODE) {
				// Block Size should be 4
				if(ext[0]==4) {
					// Read delay 
					delay = (ext[3]<<8)|ext[2];
				
					// Read disposal method
					tdisp = (ext[1]&0x1c)>>2;

					// Read transperant value, if any
					transpval = -1;
					if(ext[1]&1) {
						transpval = (int)ext[4];
					}
				}	
			}
			
			while(ext) {
				if(DGifGetExtensionNext(gif, &ext)==GIF_ERROR) {
					DGifCloseFile(gif);
					return EBADIMG;
				}
			}
		}
	} while(rec!=TERMINATE_RECORD_TYPE);

	if(row!=NULL)
		free(row);
	DGifCloseFile(gif);

	return 0;
}
