/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//(C) PicoPeta Simputers Pvt. Ltd. (2003)
// PhotoAlbum

// Programmer: Santhosh N

// TODO: Implement zoom quality

#include "plib.h"

unsigned int *build_index(unsigned int width, float zoom, unsigned int *rwidth)
{
	float fzoom;
	unsigned int *index, a;

	if(zoom==0.0) {
		fzoom = 100.0;
		*rwidth = width;
	} else {
		fzoom = zoom;
		*rwidth = (fzoom*width+0.5);
	}
	
	index = (unsigned int*)malloc(sizeof(unsigned int)*(*rwidth));
	
	for(a=0; a<*rwidth; a++)
		if(zoom)
			*(index+a) = (float)a/fzoom;
		else
			*(index+a) = a;
	
	return(index);
}

int pimg_zoom(Display *disp, struct Image **input, struct Image **output, 
											float xfact, float yfact, unsigned quality)
{
	unsigned char *ibuff;
	struct Image *img, *zimg;
	int depth, x, y;
	unsigned int *xindex, *yindex, xwidth, ywidth;

	if(input==NULL)
		return EBADIMG;

	depth = DefaultDepth(disp, DefaultScreen(disp));

	if(depth!=4 && depth!=12 && depth!=16)
		return EBADDEP;

	img = *input;
	
	if(xfact<=0.0 || yfact<=0.0)
		return EBADVAL;

	xindex = build_index(img->width, xfact, &xwidth);
	yindex = build_index(img->height, yfact, &ywidth);

	// Allocate memory for the zoomed image
	if(depth==4) {
		ibuff = (unsigned char*)malloc(xwidth*(ywidth+1));
	} else {
		ibuff = (unsigned char*)malloc(xwidth*(ywidth+1)*2);
	}
	if(ibuff==NULL) {
		free(xindex);
		free(yindex);
		return ENOMEMR;
	}

	// Create a image node
	zimg = (struct Image*)malloc(sizeof(struct Image));
     zimg->delay = 0;
     zimg->x = 0;
     zimg->y = 0;
     zimg->next = NULL;
     zimg->width = xwidth;
     zimg->height = ywidth;
     zimg->img = zimg->mimg = NULL;
		
	switch(depth) {
		case 4:
			for(y=0; y<ywidth; y++)
				for(x=0; x<xwidth; x++)
					memcpy(ibuff + y*xwidth + x, 
							img->img->data + yindex[y]*img->width + xindex[x], 1);
			break;
		case 12:
			for(y=0; y<ywidth; y++)
				for(x=0; x<xwidth; x++)
					memcpy(ibuff + y*xwidth*2 + x*2, 
							img->img->data + yindex[y]*img->width*2 + xindex[x]*2, 2);
			break;
		case 16:
			for(y=0; y<ywidth; y++)
				for(x=0; x<xwidth; x++)
					memcpy(ibuff + y*xwidth*2 + x*2, 
							img->img->data + yindex[y]*img->width*2 + xindex[x]*2, 2);
			break;
	}

	if(depth==4) {
		zimg->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
                              ibuff, xwidth, ywidth, 8, xwidth);
	} else {
		zimg->img = XCreateImage(disp, CopyFromParent, depth, ZPixmap, 0,\
                              ibuff, xwidth, ywidth, 16, xwidth*2);
	}

	*output = zimg;
	free(xindex);
	free(yindex);

	return 0;
}
